# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from unittest.mock import patch, call, DEFAULT, Mock, PropertyMock
import pytest
from functools import partial

from c7n.config import Config
from c7n_mailer.cli import CONFIG_SCHEMA as MAILER_SCHEMA

from manheim_c7n_tools.vendor.mugc import AWS
import manheim_c7n_tools.runner as runner
from manheim_c7n_tools.runner import BaseStep
from manheim_c7n_tools.utils import bold
from manheim_c7n_tools.config import ManheimConfig
from c7n_mailer.deploy import get_archive
from c7n.mu import PythonPackageArchive

pbm = 'manheim_c7n_tools.runner'


ALL_REGIONS = [
    "ap-south-1",
    "eu-west-3",
    "eu-west-2",
    "eu-west-1",
    "ap-northeast-2",
    "ap-northeast-1",
    "sa-east-1",
    "ca-central-1",
    "ap-southeast-1",
    "ap-southeast-2",
    "eu-central-1",
    "us-east-1",
    "us-east-2",
    "us-west-1",
    "us-west-2"
]


class FakeConfig:

    def __init__(self, regions):
        self.regions = regions


class StepTester(object):

    def setup(self):
        # in order to supplant __getattr__ calls
        self.m_conf = Mock(spec=ManheimConfig)
        self.m_conf.account_id = '01234567890'


class TestPolicygenStep(StepTester):

    def test_run(self):
        with patch('%s.PolicyGen' % pbm, autospec=True) as mock_pg:
            runner.PolicygenStep(None, self.m_conf).run()
        assert mock_pg.mock_calls == [
            call(self.m_conf),
            call().run()
        ]

    def test_dryrun(self):
        with patch('%s.PolicyGen' % pbm, autospec=True) as mock_pg:
            runner.PolicygenStep(None, self.m_conf).dryrun()
        assert mock_pg.mock_calls == [
            call(self.m_conf),
            call().run()
        ]

    def test_run_in_region(self):
        conf = FakeConfig(ALL_REGIONS)
        for rname in ALL_REGIONS:
            if rname == ALL_REGIONS[0]:
                assert runner.PolicygenStep.run_in_region(rname, conf) is True
            else:
                assert runner.PolicygenStep.run_in_region(rname, conf) is False


class TestValidateStep(StepTester):

    def test_run(self):
        mock_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.validate' % pbm, autospec=True) as mock_validate:
            with patch('%s.Config.empty' % pbm) as mock_empty:
                mock_empty.return_value = mock_conf
                runner.ValidateStep('rName', self.m_conf).run()
        assert mock_validate.mock_calls == [call(mock_conf)]
        assert mock_empty.mock_calls == [
            call(configs=['custodian_rName.yml'], region='rName',
                 check_deprecations="yes")
        ]

    def test_dryrun(self):
        mock_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.validate' % pbm, autospec=True) as mock_validate:
            with patch('%s.Config.empty' % pbm) as mock_empty:
                mock_empty.return_value = mock_conf
                runner.ValidateStep('rName', self.m_conf).dryrun()
        assert mock_validate.mock_calls == [call(mock_conf)]
        assert mock_empty.mock_calls == [
            call(configs=['custodian_rName.yml'], region='rName',
                 check_deprecations="yes")
        ]

    def test_run_in_region(self):
        for rname in ALL_REGIONS:
            assert runner.ValidateStep.run_in_region(rname, None) is True


class TestMugcStep(StepTester):

    def test_run(self):
        type(self.m_conf).function_prefix = PropertyMock(
            return_value='foobar-'
        )
        mock_pol1 = Mock(provider_name='foo')
        mock_pol2 = Mock(provider_name='aws')
        mock_pol3 = Mock(provider_name='azure')
        mock_conf = Mock(spec_set=ManheimConfig)
        mock_aws = Mock(spec_set=AWS)
        mock_aws.initialize_policies.return_value = {'aws': 'policies'}
        mock_pc = Mock()
        with patch('%s.Config.empty' % pbm) as mock_empty:
            mock_empty.return_value = mock_conf
            with patch.multiple(
                pbm,
                AWS=DEFAULT,
                load_policies=DEFAULT,
                resources_gc_prefix=DEFAULT,
                PolicyCollection=DEFAULT
            ) as mocks:
                mocks['AWS'].return_value = mock_aws
                mocks['PolicyCollection'].return_value = mock_pc
                mocks['load_policies'].return_value = [
                    mock_pol1, mock_pol2, mock_pol3
                ]
                runner.MugcStep('rName', self.m_conf).run()
        assert mock_empty.mock_calls == [
            call(
                config_files=['custodian_rName.yml'],
                regions=['rName'],
                prefix='foobar-',
                policy_regex='^foobar\\-.*',
                assume=None,
                policy_filter=None,
                log_group=None,
                external_id=None,
                cache_period=0,
                cache=None,
                present=False
            )
        ]
        assert mocks['load_policies'].mock_calls == [
            call(mock_conf, mock_conf)
        ]
        assert mocks['PolicyCollection'].mock_calls == [
            call([mock_pol2], mock_conf)
        ]
        assert mocks['AWS'].mock_calls == [
            call(),
            call().initialize_policies(mock_pc, mock_conf)
        ]
        assert mocks['resources_gc_prefix'].mock_calls == [
            call(mock_conf, mock_conf, {'aws': 'policies'})
        ]

    def test_dryrun(self):
        type(self.m_conf).function_prefix = PropertyMock(
            return_value='foobar-'
        )
        mock_pol1 = Mock(provider_name='foo')
        mock_pol2 = Mock(provider_name='aws')
        mock_pol3 = Mock(provider_name='azure')
        mock_conf = Mock(spec_set=ManheimConfig)
        mock_aws = Mock(spec_set=AWS)
        mock_aws.initialize_policies.return_value = {'aws': 'policies'}
        mock_pc = Mock()
        with patch('%s.Config.empty' % pbm) as mock_empty:
            mock_empty.return_value = mock_conf
            with patch.multiple(
                pbm,
                AWS=DEFAULT,
                load_policies=DEFAULT,
                resources_gc_prefix=DEFAULT,
                PolicyCollection=DEFAULT
            ) as mocks:
                mocks['AWS'].return_value = mock_aws
                mocks['PolicyCollection'].return_value = mock_pc
                mocks['load_policies'].return_value = [
                    mock_pol1, mock_pol2, mock_pol3
                ]
                runner.MugcStep('rName', self.m_conf).dryrun()
        assert mock_empty.mock_calls == [
            call(
                config_files=['custodian_rName.yml'],
                regions=['rName'],
                prefix='foobar-',
                policy_regex='^foobar\\-.*',
                assume=None,
                policy_filter=None,
                log_group=None,
                external_id=None,
                cache_period=0,
                cache=None,
                present=False,
                dryrun=True
            )
        ]
        assert mocks['load_policies'].mock_calls == [
            call(mock_conf, mock_conf)
        ]
        assert mocks['PolicyCollection'].mock_calls == [
            call([mock_pol2], mock_conf)
        ]
        assert mocks['AWS'].mock_calls == [
            call(),
            call().initialize_policies(mock_pc, mock_conf)
        ]
        assert mocks['resources_gc_prefix'].mock_calls == [
            call(mock_conf, mock_conf, {'aws': 'policies'})
        ]

    def test_run_in_region(self):
        for rname in ALL_REGIONS:
            assert runner.MugcStep.run_in_region(rname, None) is True


class TestCustodianStep(StepTester):

    def test_run(self):
        type(self.m_conf).output_s3_bucket_name = PropertyMock(
            return_value='cloud-custodian-ACCT-REGION'
        )
        type(self.m_conf).custodian_log_group = PropertyMock(
            return_value='/cloud-custodian/ACCT/REGION'
        )
        mock_conf = Mock(spec_set=Config)
        with patch('%s.run' % pbm) as mock_run:
            with patch('%s.Config.empty' % pbm) as mock_empty:
                mock_empty.return_value = mock_conf
                runner.CustodianStep('rName', self.m_conf).run()
        assert mock_run.mock_calls == [call(mock_conf)]
        assert mock_empty.mock_calls == [
            call(
                configs=['custodian_rName.yml'],
                region='rName',
                regions=['rName'],
                log_group='/cloud-custodian/ACCT/REGION',
                verbose=1,
                metrics_enabled=True,
                subparser='run',
                cache='/tmp/.cache/cloud-custodian.cache',
                command='c7n.commands.run',
                output_dir='cloud-custodian-ACCT-REGION/logs',
                vars=None,
                dryrun=False
            )
        ]

    def test_dryrun(self):
        type(self.m_conf).output_s3_bucket_name = PropertyMock(
            return_value='cloud-custodian-ACCT-REGION'
        )
        type(self.m_conf).custodian_log_group = PropertyMock(
            return_value='/cloud-custodian/ACCT/REGION'
        )
        mock_conf = Mock(spec_set=Config)
        with patch('%s.run' % pbm) as mock_run:
            with patch('%s.Config.empty' % pbm) as mock_empty:
                mock_empty.return_value = mock_conf
                runner.CustodianStep('rName', self.m_conf).dryrun()
        assert mock_run.mock_calls == [call(mock_conf)]
        assert mock_empty.mock_calls == [
            call(
                configs=['custodian_rName.yml'],
                region='rName',
                regions=['rName'],
                verbose=1,
                metrics_enabled=False,
                subparser='run',
                cache='/tmp/.cache/cloud-custodian.cache',
                command='c7n.commands.run',
                output_dir='dryrun/rName',
                vars=None,
                dryrun=True
            )
        ]

    def test_run_in_region(self):
        for rname in ALL_REGIONS:
            assert runner.CustodianStep.run_in_region(rname, None) is True


class TestMailerStep(StepTester):

    @patch(f'{pbm}.__file__', 'path/to/runner.py')
    def test_mailer_config_docker(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_config = PropertyMock(
            return_value={'mailer': 'config'}
        )

        def se_mailer_setup_defaults(d):
            d['defaults'] = 'set'

        def se_isdir(d):
            if d == '/manheim_c7n_tools/manheim_c7n_tools/mailer-templates':
                return True
            return False

        def se_abspath(p):
            return f'/abspath/{p}'

        with patch(
            '%s.jsonschema.validate' % pbm, autospec=True
        ) as mock_validate:
            with patch(
                '%s.mailer_setup_defaults' % pbm, autospec=True
            ) as mock_msd:
                mock_msd.side_effect = se_mailer_setup_defaults
                with patch(
                    '%s.os.path.isdir' % pbm
                ) as mock_isdir:
                    mock_isdir.side_effect = se_isdir
                    with patch(
                        '%s.os.path.abspath' % pbm
                    ) as mock_abspath:
                        mock_abspath.side_effect = se_abspath
                        res = runner.MailerStep(
                            'rName', m_conf
                        ).mailer_config
        expected = {
            'mailer': 'config',
            'defaults': 'set',
            'templates_folders': [
                '/manheim_c7n_tools/manheim_c7n_tools/mailer-templates'
            ]
        }
        assert res == expected
        assert mock_validate.mock_calls == [
            call(expected, MAILER_SCHEMA)
        ]
        assert mock_msd.mock_calls == [
            call(expected)
        ]

    @patch(f'{pbm}.__file__', 'path/to/runner.py')
    def test_mailer_config_nondocker(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_config = PropertyMock(
            return_value={'mailer': 'config'}
        )

        def se_mailer_setup_defaults(d):
            d['defaults'] = 'set'
            d['templates_folders'] = []

        def se_isdir(d):
            if d == '/abspath/path/to/mailer-templates':
                return True
            return False

        def se_abspath(p):
            return f'/abspath/{p}'

        with patch(
            '%s.jsonschema.validate' % pbm, autospec=True
        ) as mock_validate:
            with patch(
                '%s.mailer_setup_defaults' % pbm, autospec=True
            ) as mock_msd:
                mock_msd.side_effect = se_mailer_setup_defaults
                with patch(
                    '%s.os.path.isdir' % pbm
                ) as mock_isdir:
                    mock_isdir.side_effect = se_isdir
                    with patch(
                        '%s.os.path.abspath' % pbm
                    ) as mock_abspath:
                        mock_abspath.side_effect = se_abspath
                        res = runner.MailerStep(
                            'rName', m_conf
                        ).mailer_config
        expected = {
            'mailer': 'config',
            'defaults': 'set',
            'templates_folders': ['/abspath/path/to/mailer-templates']
        }
        assert res == expected
        assert mock_validate.mock_calls == [
            call(expected, MAILER_SCHEMA)
        ]
        assert mock_msd.mock_calls == [
            call(expected)
        ]

    @patch(f'{pbm}.__file__', 'path/to/runner.py')
    def test_mailer_config_existing_folders(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_config = PropertyMock(
            return_value={'mailer': 'config'}
        )

        def se_mailer_setup_defaults(d):
            d['defaults'] = 'set'
            d['templates_folders'] = ['/my/folder']

        def se_isdir(d):
            if d == '/abspath/path/to/mailer-templates':
                return True
            return False

        def se_abspath(p):
            return f'/abspath/{p}'

        with patch(
            '%s.jsonschema.validate' % pbm, autospec=True
        ) as mock_validate:
            with patch(
                '%s.mailer_setup_defaults' % pbm, autospec=True
            ) as mock_msd:
                mock_msd.side_effect = se_mailer_setup_defaults
                with patch(
                    '%s.os.path.isdir' % pbm
                ) as mock_isdir:
                    mock_isdir.side_effect = se_isdir
                    with patch(
                        '%s.os.path.abspath' % pbm
                    ) as mock_abspath:
                        mock_abspath.side_effect = se_abspath
                        res = runner.MailerStep(
                            'rName', m_conf
                        ).mailer_config
        expected = {
            'mailer': 'config',
            'defaults': 'set',
            'templates_folders': [
                '/abspath/path/to/mailer-templates',
                '/my/folder'
            ]
        }
        assert res == expected
        assert mock_validate.mock_calls == [
            call(expected, MAILER_SCHEMA)
        ]
        assert mock_msd.mock_calls == [
            call(expected)
        ]

    @patch(f'{pbm}.__file__', 'path/to/runner.py')
    def test_mailer_config_only_existing(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_config = PropertyMock(
            return_value={'mailer': 'config'}
        )

        def se_mailer_setup_defaults(d):
            d['defaults'] = 'set'
            d['templates_folders'] = ['/my/folder']

        def se_isdir(d):
            return False

        def se_abspath(p):
            return f'/abspath/{p}'

        with patch(
            '%s.jsonschema.validate' % pbm, autospec=True
        ) as mock_validate:
            with patch(
                '%s.mailer_setup_defaults' % pbm, autospec=True
            ) as mock_msd:
                mock_msd.side_effect = se_mailer_setup_defaults
                with patch(
                    '%s.os.path.isdir' % pbm
                ) as mock_isdir:
                    mock_isdir.side_effect = se_isdir
                    with patch(
                        '%s.os.path.abspath' % pbm
                    ) as mock_abspath:
                        mock_abspath.side_effect = se_abspath
                        res = runner.MailerStep(
                            'rName', m_conf
                        ).mailer_config
        expected = {
            'mailer': 'config',
            'defaults': 'set',
            'templates_folders': [
                '/my/folder'
            ]
        }
        assert res == expected
        assert mock_validate.mock_calls == [
            call(expected, MAILER_SCHEMA)
        ]
        assert mock_msd.mock_calls == [
            call(expected)
        ]

    @patch(f'{pbm}.__file__', 'path/to/runner.py')
    def test_mailer_config_pwd(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_config = PropertyMock(
            return_value={'mailer': 'config'}
        )

        def se_mailer_setup_defaults(d):
            d['defaults'] = 'set'

        def se_isdir(d):
            if d == '/abspath/./mailer-templates':
                return True
            return False

        def se_abspath(p):
            return f'/abspath/{p}'

        with patch(
            '%s.jsonschema.validate' % pbm, autospec=True
        ) as mock_validate:
            with patch(
                '%s.mailer_setup_defaults' % pbm, autospec=True
            ) as mock_msd:
                mock_msd.side_effect = se_mailer_setup_defaults
                with patch(
                    '%s.os.path.isdir' % pbm
                ) as mock_isdir:
                    mock_isdir.side_effect = se_isdir
                    with patch(
                        '%s.os.path.abspath' % pbm
                    ) as mock_abspath:
                        mock_abspath.side_effect = se_abspath
                        res = runner.MailerStep(
                            'rName', m_conf
                        ).mailer_config
        expected = {
            'mailer': 'config',
            'defaults': 'set',
            'templates_folders': [
                '/abspath/./mailer-templates'
            ]
        }
        assert res == expected
        assert mock_validate.mock_calls == [
            call(expected, MAILER_SCHEMA)
        ]
        assert mock_msd.mock_calls == [
            call(expected)
        ]

    @patch(f'{pbm}.__file__', 'path/to/runner.py')
    def test_mailer_config_all_exist(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_config = PropertyMock(
            return_value={'mailer': 'config'}
        )

        def se_mailer_setup_defaults(d):
            d['defaults'] = 'set'
            d['templates_folders'] = ['/my/folder']

        def se_isdir(d):
            return True

        def se_abspath(p):
            return f'/abspath/{p}'

        with patch(
            '%s.jsonschema.validate' % pbm, autospec=True
        ) as mock_validate:
            with patch(
                '%s.mailer_setup_defaults' % pbm, autospec=True
            ) as mock_msd:
                mock_msd.side_effect = se_mailer_setup_defaults
                with patch(
                    '%s.os.path.isdir' % pbm
                ) as mock_isdir:
                    mock_isdir.side_effect = se_isdir
                    with patch(
                        '%s.os.path.abspath' % pbm
                    ) as mock_abspath:
                        mock_abspath.side_effect = se_abspath
                        res = runner.MailerStep(
                            'rName', m_conf
                        ).mailer_config
        expected = {
            'mailer': 'config',
            'defaults': 'set',
            'templates_folders': [
                '/abspath/path/to/mailer-templates',
                '/manheim_c7n_tools/manheim_c7n_tools/mailer-templates',
                '/abspath/./mailer-templates',
                '/my/folder'
            ]
        }
        assert res == expected
        assert mock_validate.mock_calls == [
            call(expected, MAILER_SCHEMA)
        ]
        assert mock_msd.mock_calls == [
            call(expected)
        ]

    def test_run(self):
        m_partial = Mock(spec_set=partial)
        m_conf = Mock(spec_set=ManheimConfig)
        with patch(
            '%s.MailerStep.mailer_config' % pbm, new_callable=PropertyMock
        ) as mock_config:
            with patch(
                '%s.mailer_deploy.provision' % pbm, autospec=True
            ) as mock_prov:
                with patch(
                    '%s.session_factory' % pbm, autospec=True
                ) as mock_sf:
                    with patch(
                        '%s.functools.partial' % pbm, autospec=True
                    ) as mock_partial:
                        mock_config.return_value = m_conf
                        mock_partial.return_value = m_partial
                        runner.MailerStep('rName', self.m_conf).run()
        assert mock_partial.mock_calls == [call(mock_sf, m_conf)]
        assert mock_prov.mock_calls == [call(m_conf, m_partial)]
        assert mock_config.mock_calls == [call()]

    def test_dryrun(self):
        m_partial = Mock(spec_set=partial)
        m_conf = Mock(spec_set=ManheimConfig)
        with patch(
            '%s.MailerStep.mailer_config' % pbm, new_callable=PropertyMock
        ) as mock_config:
            with patch(
                '%s.mailer_deploy.provision' % pbm, autospec=True
            ) as mock_prov:
                with patch(
                    '%s.session_factory' % pbm, autospec=True
                ):
                    with patch(
                        '%s.functools.partial' % pbm, autospec=True
                    ) as mock_partial:
                        mock_config.return_value = m_conf
                        mock_partial.return_value = m_partial
                        runner.MailerStep('rName', self.m_conf).dryrun()
        assert mock_partial.mock_calls == []
        assert mock_prov.mock_calls == []
        assert mock_config.mock_calls == [call()]

    def test_run_in_region(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).mailer_regions = PropertyMock(
            return_value=['us-east-1']
        )
        for rname in ALL_REGIONS:
            if rname == 'us-east-1':
                assert runner.MailerStep.run_in_region(rname, m_conf) is True
            else:
                assert runner.MailerStep.run_in_region(rname, m_conf) is False

    def test_mailer_archive(self):
        """
        This is a test of ``c7n_mailer.deploy.get_archive()``. We've had a few
        dependency issues in the past that aren't detected until actually
        deploying mailer and generating the archive (zip) for the Lambda, so
        this attempts to detect those issues.
        """
        arch = get_archive({'templates_folders': []})
        assert isinstance(arch, PythonPackageArchive)
        assert arch.size > 0
        assert len(arch.get_filenames()) > 0


class TestDryRunDiffStep(StepTester):

    def test_run(self):
        with patch('%s.DryRunDiffer' % pbm, autospec=True) as mock_drd:
            runner.DryRunDiffStep('rName', self.m_conf).run()
        assert mock_drd.mock_calls == []

    def test_dryrun(self):
        with patch('%s.DryRunDiffer' % pbm, autospec=True) as mock_drd:
            runner.DryRunDiffStep('rName', self.m_conf).dryrun()
        assert mock_drd.mock_calls == [
            call(self.m_conf),
            call().run(diff_against='origin/master')
        ]

    def test_run_in_region(self):
        type(self.m_conf).regions = PropertyMock(
            return_value=ALL_REGIONS
        )
        for rname in ALL_REGIONS:
            if rname == 'us-west-2':
                assert runner.DryRunDiffStep.run_in_region(
                    rname, self.m_conf
                ) is True
            else:
                assert runner.DryRunDiffStep.run_in_region(
                    rname, self.m_conf
                ) is False


class TestS3ArchiverStep(StepTester):

    def test_run(self):
        type(self.m_conf).output_s3_bucket_name = PropertyMock(
            return_value='cloud-custodian-ACCT-REGION'
        )
        with patch('%s.S3Archiver' % pbm, autospec=True) as mock_s3a:
            runner.S3ArchiverStep('rName', self.m_conf).run()
        assert mock_s3a.mock_calls == [
            call(
                'rName',
                'cloud-custodian-ACCT-REGION',
                'custodian_rName.yml'
            ),
            call().run()
        ]

    def test_dryrun(self):
        type(self.m_conf).output_s3_bucket_name = PropertyMock(
            return_value='cloud-custodian-ACCT-REGION'
        )
        with patch('%s.S3Archiver' % pbm, autospec=True) as mock_s3a:
            runner.S3ArchiverStep('rName', self.m_conf).dryrun()
        assert mock_s3a.mock_calls == [
            call(
                'rName',
                'cloud-custodian-ACCT-REGION',
                'custodian_rName.yml',
                dryrun=True
            ),
            call().run()
        ]

    def test_run_in_region(self):
        for rname in ALL_REGIONS:
            assert runner.S3ArchiverStep.run_in_region(rname, None) is True


class TestDocsBuildStep(StepTester):

    def test_run(self):
        with patch(
            '%s.DocsBuildStep._run_sphinx_build' % pbm, autospec=True
        ) as m_rsb:
            cls = runner.DocsBuildStep('rName', self.m_conf)
            cls.run()
        assert m_rsb.mock_calls == [call(cls)]

    def test_dryrun(self):
        with patch(
            '%s.DocsBuildStep._run_sphinx_build' % pbm, autospec=True
        ) as m_rsb:
            cls = runner.DocsBuildStep('rName', self.m_conf)
            cls.dryrun()
        assert m_rsb.mock_calls == [call(cls)]

    def test_run_sphinx_build(self):
        with patch('%s.os.path.exists' % pbm, autospec=True) as mock_ope:
            with patch('%s.rmtree' % pbm, autospec=True) as mock_rmtree:
                with patch(
                    '%s.sphinx_main' % pbm, autospec=True
                ) as mock_sphinx:
                    mock_ope.return_value = False
                    mock_sphinx.return_value = 0
                    runner.DocsBuildStep(
                        'rName', self.m_conf
                    )._run_sphinx_build()
        assert mock_ope.mock_calls == [call('docs/_build')]
        assert mock_rmtree.mock_calls == []
        assert mock_sphinx.mock_calls == [
            call(['-W', 'docs/source', 'docs/_build', '-b', 'dirhtml'])
        ]

    def test_run_sphinx_build_failure(self):
        with patch('%s.os.path.exists' % pbm, autospec=True) as mock_ope:
            with patch('%s.rmtree' % pbm, autospec=True) as mock_rmtree:
                with patch(
                    '%s.sphinx_main' % pbm, autospec=True
                ) as mock_sphinx:
                    mock_ope.return_value = True
                    mock_sphinx.return_value = 3
                    with pytest.raises(RuntimeError) as exc:
                        runner.DocsBuildStep(
                            'rName', self.m_conf
                        )._run_sphinx_build()
        assert str(exc.value) == 'Sphinx exited 3'
        assert mock_ope.mock_calls == [call('docs/_build')]
        assert mock_rmtree.mock_calls == [call('docs/_build')]
        assert mock_sphinx.mock_calls == [
            call(['-W', 'docs/source', 'docs/_build', '-b', 'dirhtml'])
        ]

    def test_run_in_region(self):
        conf = FakeConfig(ALL_REGIONS)
        for rname in ALL_REGIONS:
            if rname == ALL_REGIONS[0]:
                assert runner.DocsBuildStep.run_in_region(rname, conf) is True
            else:
                assert runner.DocsBuildStep.run_in_region(rname, conf) is False


class TestStepClasses(object):

    def test_all_subclasses_have_unique_name(self):
        subc = [x for x in runner.BaseStep.__subclasses__()]
        names = []
        for x in subc:
            assert x.name is not None
            names.append(x.name.strip())
        assert len(names) == len(set(names))
        assert None not in names
        assert '' not in names

    def test_base_step_run_in_region(self):
        assert runner.BaseStep.run_in_region('foo', None) is True


class TestCustodianRunner(object):

    def setup(self):

        def se_cls2(rname, r_conf):
            return rname in ['r1', 'r3']

        def se_cls3(rname, r_conf):
            return rname == 'r1'

        self.cls1 = Mock(spec_set=BaseStep)
        type(self.cls1).name = PropertyMock(return_value='cls1')
        self.cls1.run_in_region.return_value = True
        self.cls2 = Mock(spec_set=BaseStep)
        type(self.cls2).name = PropertyMock(return_value='cls2')
        self.cls2.run_in_region.side_effect = se_cls2
        self.cls3 = Mock(spec_set=BaseStep)
        type(self.cls3).name = PropertyMock(return_value='cls3')
        self.cls3.run_in_region.side_effect = se_cls3
        self.cls4 = Mock(spec_set=BaseStep)
        type(self.cls4).name = PropertyMock(return_value='cls4')
        self.cls4.run_in_region.return_value = True
        self.steps = [self.cls1, self.cls2, self.cls3, self.cls4]

    def test_init(self):
        m_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
            mock_cff.return_value = m_conf
            cls = runner.CustodianRunner('acctName', 'cpath')
        assert cls.config == m_conf
        assert cls._config_path == 'cpath'
        assert mock_cff.mock_calls == [call('cpath', 'acctName')]

    def test_run_all_steps(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).regions = PropertyMock(
            return_value=['r1', 'r2', 'r3']
        )
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch.multiple(
                '%s.CustodianRunner' % pbm,
                autospec=True,
                _steps_to_run=DEFAULT,
                _run_step_in_regions=DEFAULT,
                _validate_account=DEFAULT
            ) as mocks:
                mocks['_steps_to_run'].return_value = [
                    self.cls1, self.cls2, self.cls3, self.cls4
                ]
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch(
                        '%s.ManheimConfig.from_file' % pbm
                    ) as mock_cff:
                        mock_cff.return_value = m_conf
                        cls = runner.CustodianRunner('acctName')
                        cls.run('run')
        assert mocks['_steps_to_run'].mock_calls == [call(cls, [], [])]
        assert mocks['_run_step_in_regions'].mock_calls == [
            call(cls, 'run', self.cls1, ['r1', 'r2', 'r3']),
            call(cls, 'run', self.cls2, ['r1', 'r2', 'r3']),
            call(cls, 'run', self.cls3, ['r1', 'r2', 'r3']),
            call(cls, 'run', self.cls4, ['r1', 'r2', 'r3'])
        ]
        assert self.cls1.mock_calls == []
        assert self.cls2.mock_calls == []
        assert self.cls3.mock_calls == []
        assert self.cls4.mock_calls == []
        assert mock_logger.mock_calls == [
            call.info(bold('Beginning run - 4 steps')),
            call.info(bold('Step 1 of 4 - cls1')),
            call.info(bold('Step 2 of 4 - cls2')),
            call.info(bold('Step 3 of 4 - cls3')),
            call.info(bold('Step 4 of 4 - cls4')),
            call.info(bold('SUCCESS: All 4 steps complete!'))
        ]
        assert mock_cff.mock_calls == [
            call('manheim-c7n-tools.yml', 'acctName')
        ]
        assert mocks['_validate_account'].mock_calls == [call(cls)]

    def test_run_dryrun_some_steps_some_regions(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).regions = PropertyMock(
            return_value=['r1', 'r2', 'r3']
        )
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch.multiple(
                '%s.CustodianRunner' % pbm,
                autospec=True,
                _steps_to_run=DEFAULT,
                _run_step_in_regions=DEFAULT,
                _validate_account=DEFAULT
            ) as mocks:
                mocks['_steps_to_run'].return_value = [
                    self.cls2, self.cls3
                ]
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                        mock_cff.return_value = m_conf
                        cls = runner.CustodianRunner('aName')
                        cls.run(
                            'dryrun',
                            regions=['r2'],
                            step_names=['cls2', 'cls3', 'cls4'],
                            skip_steps=['cls4']
                        )
        assert mocks['_steps_to_run'].mock_calls == [
            call(cls, ['cls2', 'cls3', 'cls4'], ['cls4'])
        ]
        assert mocks['_run_step_in_regions'].mock_calls == [
            call(cls, 'dryrun', self.cls2, ['r2']),
            call(cls, 'dryrun', self.cls3, ['r2'])
        ]
        assert self.cls1.mock_calls == []
        assert self.cls2.mock_calls == []
        assert self.cls3.mock_calls == []
        assert self.cls4.mock_calls == []
        assert mock_logger.mock_calls == [
            call.info(bold('Beginning dryrun - 2 of 4 steps selected')),
            call.info(bold('Step 1 of 2 - cls2')),
            call.info(bold('Step 2 of 2 - cls3')),
            call.info(bold('SUCCESS: All 2 steps complete!'))
        ]
        assert mock_cff.mock_calls == [call('manheim-c7n-tools.yml', 'aName')]
        assert mocks['_validate_account'].mock_calls == [call(cls)]

    def test_run_invalid_region_name(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).regions = PropertyMock(
            return_value=['r1', 'r2', 'r3']
        )
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch.multiple(
                '%s.CustodianRunner' % pbm,
                autospec=True,
                _steps_to_run=DEFAULT,
                _run_step_in_regions=DEFAULT,
                _validate_account=DEFAULT
            ) as mocks:
                mocks['_steps_to_run'].return_value = [
                    self.cls2, self.cls3
                ]
                with patch('%s.logger' % pbm, autospec=True) as mock_logger:
                    with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                        mock_cff.return_value = m_conf
                        with pytest.raises(RuntimeError) as exc:
                            cls = runner.CustodianRunner('acctName')
                            cls.run(
                                'dryrun',
                                regions=['notValid'],
                                step_names=['cls2', 'cls3', 'cls4'],
                                skip_steps=['cls4']
                            )
        assert str(exc.value) == 'ERROR: All specified region names must be ' \
                                 'listed in the "regions" section of the ' \
                                 'config file (manheim-c7n-tools.yml)'
        assert mocks['_steps_to_run'].mock_calls == [
            call(cls, ['cls2', 'cls3', 'cls4'], ['cls4'])
        ]
        assert mocks['_run_step_in_regions'].mock_calls == []
        assert self.cls1.mock_calls == []
        assert self.cls2.mock_calls == []
        assert self.cls3.mock_calls == []
        assert self.cls4.mock_calls == []
        assert mock_logger.mock_calls == [
            call.info(bold('Beginning dryrun - 2 of 4 steps selected'))
        ]
        assert mocks['_validate_account'].mock_calls == [call(cls)]

    def test_validate_account(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).account_name = PropertyMock(
            return_value='myAcct'
        )
        type(m_conf).account_id = PropertyMock(return_value='0234567890')

        with patch('%s.boto3.client' % pbm) as mock_client:
            mock_client.return_value.get_caller_identity.return_value = {
                'UserId': 'MyUID',
                'Arn': 'myARN',
                'Account': '0234567890'
            }
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                cls = runner.CustodianRunner('acctName')
                cls._validate_account()
        assert mock_cff.mock_calls == [
            call('manheim-c7n-tools.yml', 'acctName')
        ]
        assert mock_client.mock_calls == [
            call('sts', region_name='us-east-1'),
            call().get_caller_identity()
        ]

    def test_validate_account_failed(self):
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_conf).account_name = PropertyMock(
            return_value='myAcct'
        )
        type(m_conf).account_id = PropertyMock(return_value='1234567890')

        with patch('%s.boto3.client' % pbm) as mock_client:
            mock_client.return_value.get_caller_identity.return_value = {
                'UserId': 'MyUID',
                'Arn': 'myARN',
                'Account': '9876543210'
            }
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                cls = runner.CustodianRunner('acctName')
                with pytest.raises(RuntimeError) as exc:
                    cls._validate_account()
        assert str(exc.value) == 'ERROR: Using configuration for account ' \
                                 '1234567890 (myAcct), but ' \
                                 'sts:GetCallerIdentity reports connected to ' \
                                 'account 9876543210'
        assert mock_cff.mock_calls == [
            call('manheim-c7n-tools.yml', 'acctName')
        ]
        assert mock_client.mock_calls == [
            call('sts', region_name='us-east-1'),
            call().get_caller_identity()
        ]

    def test_steps_to_run_all(self):
        m_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                res = runner.CustodianRunner('acctName')._steps_to_run([], [])
        assert res == self.steps

    def test_steps_to_run_step_names(self):
        m_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                res = runner.CustodianRunner('acctName')._steps_to_run(
                    ['cls3', 'cls1'], []
                )
        assert res == [self.cls1, self.cls3]

    def test_steps_to_run_skip_steps(self):
        m_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                res = runner.CustodianRunner('aName')._steps_to_run(
                    [], ['cls4', 'cls2']
                )
        assert res == [self.cls1, self.cls3]

    def test_steps_to_run_names_and_skip(self):
        m_conf = Mock(spec_set=ManheimConfig)
        with patch('%s.CustodianRunner.ordered_step_classes' % pbm, self.steps):
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                res = runner.CustodianRunner('aName')._steps_to_run(
                    ['cls3', 'cls2', 'cls1'], ['cls1']
                )
        assert res == [self.cls2, self.cls3]

    def test_ordered_step_classes(self):
        """ensures all are subclasses of BaseStep"""
        m_conf = Mock(spec_set=ManheimConfig)
        for klass in runner.CustodianRunner.ordered_step_classes:
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                assert isinstance(klass(None, m_conf), runner.BaseStep)

    def test_run_in_regions_run(self):
        m_conf = Mock(spec_set=ManheimConfig)
        m_conf_r1 = Mock(spec_set=ManheimConfig)
        m_conf_r2 = Mock(spec_set=ManheimConfig)
        m_conf_r3 = Mock(spec_set=ManheimConfig)

        def se_conf_for_region(rname):
            if rname == 'r1':
                return m_conf_r1
            if rname == 'r2':
                return m_conf_r2
            if rname == 'r3':
                return m_conf_r3

        m_conf.config_for_region.side_effect = se_conf_for_region

        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                runner.CustodianRunner('acctName')._run_step_in_regions(
                    'run', self.cls1, ['r1', 'r2', 'r3']
                )
        assert self.cls1.mock_calls == [
            call.run_in_region('r1', m_conf_r1),
            call('r1', m_conf_r1),
            call().run(),
            call.run_in_region('r2', m_conf_r2),
            call('r2', m_conf_r2),
            call().run(),
            call.run_in_region('r3', m_conf_r3),
            call('r3', m_conf_r3),
            call().run()
        ]
        assert m_conf.config_for_region.mock_calls == [
            call('r1'),
            call('r2'),
            call('r3')
        ]
        assert mock_logger.mock_calls == [
            call.info(bold('Step cls1 in REGION 1 of 3 (r1)')),
            call.info(bold('Step cls1 in REGION 2 of 3 (r2)')),
            call.info(bold('Step cls1 in REGION 3 of 3 (r3)'))
        ]

    def test_run_in_regions_policygen_run(self):
        m_conf = Mock(spec_set=ManheimConfig)
        m_conf_r1 = Mock(spec_set=ManheimConfig)
        m_conf_r2 = Mock(spec_set=ManheimConfig)
        m_conf_r3 = Mock(spec_set=ManheimConfig)

        def se_conf_for_region(rname):
            if rname == 'r1':
                return m_conf_r1
            if rname == 'r2':
                return m_conf_r2
            if rname == 'r3':
                return m_conf_r3

        m_conf.config_for_region.side_effect = se_conf_for_region

        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                with patch('%s.PolicygenStep' % pbm, autospec=True) as mock_pgs:
                    type(mock_pgs).name = PropertyMock(return_value='policygen')
                    mock_pgs.run_in_region.return_value = True
                    runner.CustodianRunner('acctName')._run_step_in_regions(
                        'run', mock_pgs, ['r1', 'r2', 'r3']
                    )
        assert mock_pgs.mock_calls == [
            call.run_in_region('r1', m_conf),
            call('r1', m_conf),
            call().run(),
            call.run_in_region('r2', m_conf),
            call('r2', m_conf),
            call().run(),
            call.run_in_region('r3', m_conf),
            call('r3', m_conf),
            call().run()
        ]
        assert m_conf.config_for_region.mock_calls == []
        assert mock_logger.mock_calls == [
            call.info(bold('Step policygen in REGION 1 of 3 (r1)')),
            call.info(bold('Step policygen in REGION 2 of 3 (r2)')),
            call.info(bold('Step policygen in REGION 3 of 3 (r3)'))
        ]

    def test_run_in_regions_dryrun_skip_some(self):
        m_conf = Mock(spec_set=ManheimConfig)
        m_conf_r1 = Mock(spec_set=ManheimConfig)
        m_conf_r2 = Mock(spec_set=ManheimConfig)
        m_conf_r3 = Mock(spec_set=ManheimConfig)

        def se_conf_for_region(rname):
            if rname == 'r1':
                return m_conf_r1
            if rname == 'r2':
                return m_conf_r2
            if rname == 'r3':
                return m_conf_r3

        m_conf.config_for_region.side_effect = se_conf_for_region

        with patch('%s.logger' % pbm, autospec=True) as mock_logger:
            with patch('%s.ManheimConfig.from_file' % pbm) as mock_cff:
                mock_cff.return_value = m_conf
                runner.CustodianRunner('acctName')._run_step_in_regions(
                    'dryrun', self.cls2, ['r2', 'r3']
                )
        assert self.cls2.mock_calls == [
            call.run_in_region('r2', m_conf_r2),
            call.run_in_region('r3', m_conf_r3),
            call('r3', m_conf_r3),
            call().dryrun()
        ]
        assert m_conf.config_for_region.mock_calls == [
            call('r2'),
            call('r3')
        ]
        assert mock_logger.mock_calls == [
            call.info(bold('SKIPPING Step cls2 in REGION 1 of 2 (r2)')),
            call.info(bold('Step cls2 in REGION 2 of 2 (r3)'))
        ]


class TestParseArgs(object):

    def test_run(self):
        p = runner.parse_args(['run', 'aName'])
        assert p.verbose == 0
        assert p.steps == []
        assert p.skip == []
        assert p.ACTION == 'run'
        assert p.regions == []
        assert p.config == 'manheim-c7n-tools.yml'
        assert p.ACCT_NAME == 'aName'
        assert p.assume_role is True

    def test_run_skip_steps(self):
        p = runner.parse_args(
            ['-S', 'foo', '--skip-step=bar', 'run', 'acctName']
        )
        assert p.verbose == 0
        assert p.steps == []
        assert p.skip == ['foo', 'bar']
        assert p.ACTION == 'run'
        assert p.regions == []
        assert p.config == 'manheim-c7n-tools.yml'
        assert p.ACCT_NAME == 'acctName'
        assert p.assume_role is True

    def test_dryrun_info_region(self):
        p = runner.parse_args(['-v', '-r', 'us-east-1', 'dryrun', 'aName'])
        assert p.verbose == 1
        assert p.steps == []
        assert p.skip == []
        assert p.ACTION == 'dryrun'
        assert p.regions == ['us-east-1']
        assert p.config == 'manheim-c7n-tools.yml'
        assert p.ACCT_NAME == 'aName'
        assert p.assume_role is True

    def test_list(self):
        p = runner.parse_args(['-c', 'foobar.yml', '--no-assume-role', 'list'])
        assert p.verbose == 0
        assert p.steps == []
        assert p.skip == []
        assert p.ACTION == 'list'
        assert p.regions == []
        assert p.config == 'foobar.yml'
        assert p.assume_role is False

    def test_list_accounts(self):
        p = runner.parse_args(['accounts'])
        assert p.verbose == 0
        assert p.steps == []
        assert p.skip == []
        assert p.ACTION == 'accounts'
        assert p.regions == []
        assert p.config == 'manheim-c7n-tools.yml'
        assert p.assume_role is True

    def test_run_debug_steps_assume_role(self):
        p = runner.parse_args(
            ['-vv', '-A', '-s', 'foo', '--step=bar', 'run', 'aName']
        )
        assert p.verbose == 2
        assert p.steps == ['foo', 'bar']
        assert p.skip == []
        assert p.ACTION == 'run'
        assert p.regions == []
        assert p.config == 'manheim-c7n-tools.yml'
        assert p.ACCT_NAME == 'aName'
        assert p.assume_role is False


class FakeArgs(object):
    verbose = 0
    list = False
    steps = []
    skip = []
    ACTION = None
    regions = []
    config = 'manheim-c7n-tools.yml'
    ACCT_NAME = 'acctName'
    assume_role = True

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class TestMain(object):

    def test_run(self, capsys):
        m_cr = Mock(spec_set=runner.CustodianRunner)
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_cr).config = m_conf
        with patch.multiple(
            pbm,
            autospec=True,
            parse_args=DEFAULT,
            set_log_debug=DEFAULT,
            set_log_info=DEFAULT,
            CustodianRunner=DEFAULT,
            ManheimConfig=DEFAULT,
            assume_role=DEFAULT
        ) as mocks:
            mocks['parse_args'].return_value = FakeArgs(
                ACTION='run', regions=['foo2'], assume_role=False
            )
            mocks['CustodianRunner'].return_value = m_cr
            runner.main()
        captured = capsys.readouterr()
        assert mocks['parse_args'].mock_calls == [
            call(sys.argv[1:])
        ]
        assert mocks['set_log_debug'].mock_calls == []
        assert mocks['set_log_info'].mock_calls == []
        assert mocks['CustodianRunner'].mock_calls == [
            call('acctName', 'manheim-c7n-tools.yml'),
            call().run(
                'run', ['foo2'], step_names=[], skip_steps=[]
            )
        ]
        assert mocks['ManheimConfig'].mock_calls == []
        assert captured.out == ''
        assert captured.err == ''
        assert mocks['assume_role'].mock_calls == []

    def test_info_list(self, capsys):
        osc = runner.CustodianRunner.ordered_step_classes
        m_cr = Mock(spec_set=runner.CustodianRunner)
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_cr).config = m_conf
        with patch.multiple(
            pbm,
            autospec=True,
            parse_args=DEFAULT,
            set_log_debug=DEFAULT,
            set_log_info=DEFAULT,
            CustodianRunner=DEFAULT,
            ManheimConfig=DEFAULT,
            assume_role=DEFAULT
        ) as mocks:
            mocks['parse_args'].return_value = FakeArgs(
                ACTION='list', verbose=1
            )
            mocks['CustodianRunner'].return_value = m_cr
            type(mocks['CustodianRunner']).ordered_step_classes = osc
            with pytest.raises(SystemExit) as exc:
                runner.main()
        assert exc.value.code == 0
        captured = capsys.readouterr()
        assert mocks['parse_args'].mock_calls == [
            call(sys.argv[1:])
        ]
        assert mocks['set_log_debug'].mock_calls == []
        assert mocks['set_log_info'].mock_calls == [call(runner.logger)]
        assert mocks['ManheimConfig'].mock_calls == []
        expected = "\n".join(
            x.name for x in osc
        )
        assert captured.out == expected + "\n"
        assert captured.err == ''
        assert mocks['assume_role'].mock_calls == []

    def test_accounts(self, capsys):
        m_conf = Mock(spec_set=ManheimConfig)
        osc = runner.CustodianRunner.ordered_step_classes
        m_cr = Mock(spec_set=runner.CustodianRunner)
        type(m_cr).config = m_conf
        with patch.multiple(
            pbm,
            autospec=True,
            parse_args=DEFAULT,
            set_log_debug=DEFAULT,
            set_log_info=DEFAULT,
            CustodianRunner=DEFAULT,
            ManheimConfig=DEFAULT,
            assume_role=DEFAULT
        ) as mocks:
            mocks['parse_args'].return_value = FakeArgs(ACTION='accounts')
            mocks['CustodianRunner'].return_value = m_cr
            type(mocks['CustodianRunner']).ordered_step_classes = osc
            type(mocks['ManheimConfig']).return_value = m_conf
            with patch('%s.ManheimConfig.list_accounts' % pbm) as mock_la:
                mock_la.return_value = {
                    'acct1': 1111,
                    'acct3': 3333,
                    'acct2': 2222
                }
                with pytest.raises(SystemExit) as exc:
                    runner.main()
        assert exc.value.code == 0
        captured = capsys.readouterr()
        assert mocks['parse_args'].mock_calls == [
            call(sys.argv[1:])
        ]
        assert mocks['set_log_debug'].mock_calls == []
        assert mocks['set_log_info'].mock_calls == []
        assert mocks['ManheimConfig'].mock_calls == []
        assert mock_la.mock_calls == [call('manheim-c7n-tools.yml')]
        assert captured.out == "acct1 (1111)\nacct2 (2222)\nacct3 (3333)\n"
        assert captured.err == ''
        assert mocks['assume_role'].mock_calls == []

    def test_debug_dryrun_assume_role(self, capsys):
        m_cr = Mock(spec_set=runner.CustodianRunner)
        m_conf = Mock(spec_set=ManheimConfig)
        type(m_cr).config = m_conf
        with patch.multiple(
            pbm,
            autospec=True,
            parse_args=DEFAULT,
            set_log_debug=DEFAULT,
            set_log_info=DEFAULT,
            CustodianRunner=DEFAULT,
            ManheimConfig=DEFAULT,
            assume_role=DEFAULT
        ) as mocks:
            mocks['parse_args'].return_value = FakeArgs(
                ACTION='dryrun', verbose=2, steps=['foo'], skip=['bar'],
                config='foo.yml', ACCT_NAME='aName', assume_role=True
            )
            mocks['CustodianRunner'].return_value = m_cr
            runner.main()
        captured = capsys.readouterr()
        assert mocks['parse_args'].mock_calls == [
            call(sys.argv[1:])
        ]
        assert mocks['set_log_debug'].mock_calls == [call(runner.logger)]
        assert mocks['set_log_info'].mock_calls == []
        assert mocks['CustodianRunner'].mock_calls == [
            call('aName', 'foo.yml'),
            call().run(
                'dryrun', [], step_names=['foo'], skip_steps=['bar']
            )
        ]
        assert mocks['ManheimConfig'].mock_calls == []
        assert captured.out == ''
        assert captured.err == ''
        assert mocks['assume_role'].mock_calls == [call(m_conf)]



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command-line entrypoint to run one/multiple/all steps of a c7n deployment.

The command-line entrypoint function (:py:func:`~.main`) instantiates an
instance of :py:class:`~.CustodianRunner` and calls its
:py:meth:`~.CustodianRunner.run` method. This iterates through all of the
classes listed in :py:attr:`~.CustodianRunner.ordered_step_classes` and calls
their ``run`` or ``dryrun`` methods depending on which was specified on the
command line.
"""

import sys
import logging
import argparse
import abc
import functools
from shutil import rmtree
import os
from copy import deepcopy
import re

from sphinx.cmd.build import main as sphinx_main
import jsonschema
import boto3

from c7n.commands import validate, run
from c7n.config import Config
from c7n.policy import PolicyCollection
from c7n_mailer.cli import session_factory
from c7n_mailer.cli import CONFIG_SCHEMA as MAILER_SCHEMA
from c7n_mailer.utils import setup_defaults as mailer_setup_defaults
from c7n_mailer import deploy as mailer_deploy

from manheim_c7n_tools.utils import (
    set_log_info, set_log_debug, bold, assume_role
)
from manheim_c7n_tools.version import VERSION, PROJECT_URL
from manheim_c7n_tools.policygen import PolicyGen
from manheim_c7n_tools.vendor.mugc import (
    load_policies, resources_gc_prefix, AWS
)
from manheim_c7n_tools.dryrun_diff import DryRunDiffer
from manheim_c7n_tools.s3_archiver import S3Archiver
from manheim_c7n_tools.config import ManheimConfig

FORMAT = "[%(asctime)s %(levelname)s] %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger()

for lname in ['boto3', 'botocore', 'urllib', 'urllib3']:
    # suppress library logging below WARNING level
    log = logging.getLogger(lname)
    log.setLevel(logging.WARNING)
    log.propagate = True


class BaseStep(object):
    """
    Base class representing one step in the deployment process. Subclass this
    for each step. It should have a name and two public methods, ``run`` and
    ``dryrun``.
    """

    __metaclass__ = abc.ABCMeta

    #: The name of the step, as used on the CLI
    name = None

    def __init__(self, region_name, config):
        """
        Base Step class initializer.

        Steps should ONLY ever be initialized by
        :py:meth:`~.CustodianRunner._run_step_in_regions`.

        :param region_name: region name to run this step against
        :type region_name: str
        :param config: The manheim-c7n-tools config to use for this step,
          already made region-specific
          (:py:meth:`~.ManheimConfig.config_for_region` is called where this
          class is intialized in
          :py:meth:`~.CustodianRunner._run_step_in_regions`).
        :type config: ManheimConfig
        """
        self.region_name = region_name
        self.config = config

    @abc.abstractmethod
    def run(self):
        pass  # nocoverage

    @abc.abstractmethod
    def dryrun(self):
        pass  # nocoverage

    @staticmethod
    def run_in_region(region_name, config):
        """
        Return True if this step should run in the specified region,
        False if it should not.

        :param region_name: region name to test
        :type region_name: str
        :param config: The manheim-c7n-tools config to use for this step,
          already made region-specific
          (:py:meth:`~.ManheimConfig.config_for_region` is called where this
          class is intialized in
          :py:meth:`~.CustodianRunner._run_step_in_regions`).
        :type config: ManheimConfig
        :return: whether this step should run in the specified region
        :rtype: bool
        """
        return True


class PolicygenStep(BaseStep):
    """Step to run policygen to generate custodian-ready policies on disk."""

    name = 'policygen'

    def _do_policygen(self):
        PolicyGen(self.config).run()

    def run(self):
        self._do_policygen()

    def dryrun(self):
        self._do_policygen()

    @staticmethod
    def run_in_region(region_name, conf):
        # only run in the first-configured region
        return region_name == conf.regions[0]


class ValidateStep(BaseStep):
    """Step to run custodian validate on generated policies."""

    name = 'validate'

    def _do_validate(self):
        conf = Config.empty(
            configs=['custodian_%s.yml' % self.region_name],
            region=self.region_name,
            check_deprecations="yes"
        )
        validate(conf)

    def run(self):
        self._do_validate()

    def dryrun(self):
        self._do_validate()


class MugcStep(BaseStep):
    """
    Step to run custodian mugc.py (lambda garbage collection), based on main()
    in that module.
    """

    name = 'mugc'

    def run(self):
        # This is largely based off of mugc.main()
        logging.getLogger('botocore').setLevel(logging.ERROR)
        logging.getLogger('urllib3').setLevel(logging.ERROR)
        logging.getLogger('c7n.cache').setLevel(logging.WARNING)
        conf = Config.empty(
            config_files=['custodian_%s.yml' % self.region_name],
            regions=[self.region_name],
            prefix=self.config.function_prefix,
            policy_regex='^' + re.escape(self.config.function_prefix) + '.*',
            assume=None,
            policy_filter=None,
            log_group=None,
            external_id=None,
            cache_period=0,
            cache=None,
            present=False
        )
        # use cloud provider to initialize policies to get region expansion
        policies = AWS().initialize_policies(
            PolicyCollection(
                [
                    p for p in load_policies(conf, conf)
                    if p.provider_name == 'aws'
                ],
                conf
            ),
            conf
        )
        resources_gc_prefix(conf, conf, policies)

    def dryrun(self):
        # This is largely based off of mugc.main()
        logging.getLogger('botocore').setLevel(logging.ERROR)
        logging.getLogger('urllib3').setLevel(logging.ERROR)
        logging.getLogger('c7n.cache').setLevel(logging.WARNING)
        conf = Config.empty(
            config_files=['custodian_%s.yml' % self.region_name],
            regions=[self.region_name],
            prefix=self.config.function_prefix,
            policy_regex='^' + re.escape(self.config.function_prefix) + '.*',
            assume=None,
            policy_filter=None,
            log_group=None,
            external_id=None,
            cache_period=0,
            cache=None,
            present=False,
            dryrun=True
        )
        # use cloud provider to initialize policies to get region expansion
        policies = AWS().initialize_policies(
            PolicyCollection(
                [
                    p for p in load_policies(conf, conf)
                    if p.provider_name == 'aws'
                ],
                conf
            ),
            conf
        )
        resources_gc_prefix(conf, conf, policies)


class CustodianStep(BaseStep):
    """Step for actual custodian run"""

    name = 'custodian'

    def run(self):
        """
        Perform an actual run of cloud-custodian.

        This replicates the command:
        custodian run --region '${region}' --metrics -v -s \
          cloud-custodian-${account_id}-${region}/logs \
          --log-group=/cloud-custodian/${account_id}/${region} \
          -c custodian_${region}.yml \
          --cache '/tmp/.cache/cloud-custodian.cache'
        """
        conf = Config.empty(
            configs=['custodian_%s.yml' % self.region_name],
            region=self.region_name,
            regions=[self.region_name],
            log_group=self.config.custodian_log_group,
            verbose=1,
            metrics_enabled=True,
            subparser='run',
            cache='/tmp/.cache/cloud-custodian.cache',
            command='c7n.commands.run',
            output_dir='%s/logs' % self.config.output_s3_bucket_name,
            vars=None,
            dryrun=False
        )
        run(conf)

    def dryrun(self):
        """
        Perform a dry-run of custodian.

        This replicates the command:

        custodian run --region '${region}' --dryrun -v -s dryrun/${region} \
          -c custodian_${region}.yml \
          --cache '/tmp/.cache/cloud-custodian.cache'
        """
        conf = Config.empty(
            configs=['custodian_%s.yml' % self.region_name],
            region=self.region_name,
            regions=[self.region_name],
            verbose=1,
            metrics_enabled=False,
            subparser='run',
            cache='/tmp/.cache/cloud-custodian.cache',
            command='c7n.commands.run',
            output_dir='dryrun/%s' % self.region_name,
            vars=None,
            dryrun=True
        )
        run(conf)


class MailerStep(BaseStep):
    """
    Step for running c7n-mailer dryrun or Lambda provision

    This replicates the parts of c7n_mailer.cli that we need for our use case.
    """

    name = 'mailer'

    @property
    def mailer_config(self):
        """
        Return the validated c7n-mailer config.

        :return: c7n-mailer config
        """
        conf = deepcopy(self.config.mailer_config)
        jsonschema.validate(conf, MAILER_SCHEMA)
        mailer_setup_defaults(conf)
        logger.debug('Default mailer config: %s', conf)
        if 'templates_folders' not in conf:
            conf['templates_folders'] = []
        for d in [
            os.path.abspath('./mailer-templates'),
            '/manheim_c7n_tools/manheim_c7n_tools/mailer-templates',
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'mailer-templates'
            )
        ]:
            if os.path.isdir(d):
                conf['templates_folders'].insert(0, d)
        logger.info('Generated mailer config: %s', conf)
        return conf

    def run(self):
        conf = self.mailer_config
        mailer_deploy.provision(
            conf,
            functools.partial(session_factory, conf)
        )

    def dryrun(self):
        # The only dryrun that mailer has right now is config validation
        self.mailer_config

    @staticmethod
    def run_in_region(region_name, config):
        return region_name in config.mailer_regions


class DryRunDiffStep(BaseStep):
    """Generates the dryrun diff during dry runs."""

    name = 'dryrun-diff'

    def run(self):
        logger.info('Nothing to do during normal run.')

    def dryrun(self):
        DryRunDiffer(self.config).run(diff_against='origin/master')

    @staticmethod
    def run_in_region(region_name, conf):
        return region_name == conf.regions[-1]


class S3ArchiverStep(BaseStep):
    """Runs s3archiver to archive logs of deleted policies."""

    name = 's3archiver'

    def run(self):
        S3Archiver(
            self.region_name,
            self.config.output_s3_bucket_name,
            'custodian_%s.yml' % self.region_name
        ).run()

    def dryrun(self):
        S3Archiver(
            self.region_name,
            self.config.output_s3_bucket_name,
            'custodian_%s.yml' % self.region_name,
            dryrun=True
        ).run()


class DocsBuildStep(BaseStep):
    """Builds generated documentation."""

    name = 'docs'

    def _run_sphinx_build(self):
        if os.path.exists('docs/_build'):
            logger.info('Removing docs/_build')
            rmtree('docs/_build')
        # "sphinx-build -W docs/source docs/_build -b dirhtml"
        argv = ['-W', 'docs/source', 'docs/_build', '-b', 'dirhtml']
        logger.info('Running: sphinx-build %s' % ' '.join(argv))
        rcode = sphinx_main(argv)
        if rcode != 0:
            raise RuntimeError('Sphinx exited %d' % rcode)

    def run(self):
        self._run_sphinx_build()

    def dryrun(self):
        self._run_sphinx_build()

    @staticmethod
    def run_in_region(region_name, conf):
        # only run in the first-configured region
        return region_name == conf.regions[0]


class CustodianRunner(object):
    """
    Main class to run all steps required for manheim c7n deployment.
    """

    #: List of the :py:class:`~.BaseStep` subclasses to run for deployment,
    #: in the order they should be run.
    ordered_step_classes = [
        PolicygenStep,
        ValidateStep,
        MugcStep,
        CustodianStep,
        MailerStep,
        DryRunDiffStep,
        S3ArchiverStep,
        DocsBuildStep
    ]

    def __init__(self, account_name, config_path='manheim-c7n-tools.yml'):
        """
        Initialize the Runner.

        :param account_name: name of the account to run against
        :type account_name: str
        :param config_path: path to ``manheim-c7n-tools.yml`` config file
        :type config_path: str
        """
        self._config_path = config_path
        self.config = ManheimConfig.from_file(config_path, account_name)

    def _steps_to_run(self, step_names, skip_steps):
        """
        Given the ``step_names`` and ``skip_steps`` passed to :py:meth:`~.run`,
        return the list of step classes to run (in order).

        :param step_names: list of step names to run
        :type step_names: list
        :param skip_steps: list of step names to skip
        :type skip_steps: list
        :return: list of step classes to run, in order
        :rtype: list
        """
        if not step_names and not skip_steps:
            # both lists are empty, run everything
            return self.ordered_step_classes
        if not step_names:
            # we only had skip_steps, so start with the list of all steps
            step_names = [x.name for x in self.ordered_step_classes]
        return [
            x for x in self.ordered_step_classes
            if x.name in step_names and x.name not in skip_steps
        ]

    def run(self, action, regions=[], step_names=[], skip_steps=[]):
        """
        Main method to run all steps. This calls :py:meth:`~._steps_to_run`
        to determine which step classes to run and the order to run them in,
        and then loops through that list calling the :py:meth:`~.BaseStep.run`
        or :py:meth:`~.BaseStep.dryrun` method on each of them, according to the
        ``action`` specified.

        :param action: Name of the action to do, "run" or "dryrun"
        :type action: str
        :param regions: list of string region names to run in; if left empty,
          run in all regions listed in config file
        :type regions: list
        :param step_names: list of string step names to run; if not specified,
          will run all defined steps. Steps are always run in the order defined
          in :py:attr:`~.ordered_step_classes`.
        :type step_names: list
        :param skip_steps: list of string step names to skip running
        :type skip_steps: list
        """
        self._validate_account()
        to_run = self._steps_to_run(step_names, skip_steps)
        if to_run == self.ordered_step_classes:
            logger.info(bold(
                'Beginning %s - %d steps' % (action, len(to_run))
            ))
        else:
            logger.info(bold(
                'Beginning %s - %d of %d steps selected' % (
                    action, len(to_run), len(self.ordered_step_classes)
                )
            ))
        if regions:
            if not set(regions).issubset(set(self.config.regions)):
                raise RuntimeError(
                    'ERROR: All specified region names must be listed in the '
                    '"regions" section of the config file '
                    '(%s)' % self._config_path
                )
        else:
            # use all regions from config file
            regions = self.config.regions
        for idx, step in enumerate(to_run):
            logger.info(bold(
                'Step %d of %d - %s' % (idx + 1, len(to_run), step.name)
            ))
            self._run_step_in_regions(action, step, regions)
        logger.info(bold('SUCCESS: All %d steps complete!' % len(to_run)))

    def _validate_account(self):
        """
        Validate that we are connected to the configured account.

        :raises: RuntimeError
        """
        logger.debug('Connecting to STS in us-east-1 to verify account')
        sts = boto3.client('sts', region_name='us-east-1')
        cid = sts.get_caller_identity()
        logger.debug('Caller Identity: %s', cid)
        if cid['Account'] != self.config.account_id:
            raise RuntimeError(
                'ERROR: Using configuration for account %s (%s), but '
                'sts:GetCallerIdentity reports connected to account %s' % (
                    self.config.account_id, self.config.account_name,
                    cid['Account']
                )
            )

    def _run_step_in_regions(self, action, step, regions):
        """
        Called from :py:meth:`~.run`; run a given step in all applicable /
        specified regions.

        :param action: Name of the action to do, "run" or "dryrun"
        :type action: str
        :param step: A reference to the :py:class:`~.BaseStep` subclass to run
        :type step: object
        :param regions: list of string region names to run in
        :type regions: list
        """
        for r_idx, region_name in enumerate(regions):
            if step.name in ['policygen', 'dryrun-diff']:
                # Some steps need a config with %%AWS_REGION%% un-interpolated
                region_conf = self.config
            else:
                region_conf = self.config.config_for_region(region_name)
            if not step.run_in_region(region_name, region_conf):
                logger.info(bold(
                    'SKIPPING Step %s in REGION %d of %d (%s)' % (
                        step.name, r_idx + 1, len(regions), region_name
                    )
                ))
                continue
            logger.info(bold(
                'Step %s in REGION %d of %d (%s)' % (
                    step.name, r_idx + 1, len(regions), region_name
                )
            ))
            if action == 'run':
                step(region_name, region_conf).run()
            else:
                step(region_name, region_conf).dryrun()
            sys.stdout.flush()
            sys.stderr.flush()


def parse_args(argv):
    """Parse command-line arguments with ArgumentParser."""
    p = argparse.ArgumentParser(
        description='manheim-c7n-tools runner; runs one, multiple, or all '
                    'actions for c7n build/deploy',
        epilog='This tool is part of manheim_c7n_tools v%s.\n'
               'For documentation, see: %s' % (VERSION, PROJECT_URL),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-V', '--version', action='version', version=VERSION)
    p.add_argument('-r', '--region', dest='regions', action='append',
                   default=[],
                   help='Region names to run against; may be specified multiple'
                        'times.')
    p.add_argument('-c', '--config', dest='config', action='store',
                   default='manheim-c7n-tools.yml',
                   help='Config file path (default: ./manheim-c7n-tools.yml)')
    p.add_argument('-s', '--step', dest='steps', action='append', default=[],
                   help='Specify one or more steps to run. If not given, will '
                        'run all steps.')
    p.add_argument('-S', '--skip-step', dest='skip', action='append',
                   default=[], help='Specify one or more step names to skip.')
    p.add_argument('-A', '--no-assume-role', dest='assume_role',
                   action='store_false', default=True,
                   help='Do not assume a role, even if  specified in the '
                        'configuration file.')
    subp = p.add_subparsers(help='command', title='subcommands')

    run_parser = subp.add_parser(
        'run', help='Perform a full run (must specify ACCT_NAME)'
    )
    run_parser.set_defaults(ACTION='run')
    dryrun_parser = subp.add_parser(
        'dryrun', help='Perform a dry run (must specify ACCT_NAME)'
    )
    dryrun_parser.set_defaults(ACTION='dryrun')
    list_parser = subp.add_parser('list', help='List available steps')
    list_parser.set_defaults(ACTION='list')
    acct_parser = subp.add_parser('accounts', help='List configured accounts')
    acct_parser.set_defaults(ACTION='accounts')

    for parser in [run_parser, dryrun_parser]:
        parser.add_argument(
            'ACCT_NAME', action='store', type=str, default=None,
            help='account_name value from config file, for account to run '
                 'against'
        )

    args = p.parse_args(argv)
    return args


def main():
    """main command-line entrypoint; calls parse_args, sets up logging, and
    either lists steps or instantiates a CustodianRunner and calls run()."""
    args = parse_args(sys.argv[1:])

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    if args.ACTION == 'list':
        for x in CustodianRunner.ordered_step_classes:
            print(x.name)
        raise SystemExit(0)
    if args.ACTION == 'accounts':
        accts = ManheimConfig.list_accounts(args.config)
        for acctname in sorted(accts.keys()):
            print("%s (%s)" % (acctname, accts[acctname]))
        raise SystemExit(0)
    cr = CustodianRunner(args.ACCT_NAME, args.config)
    if args.assume_role:
        assume_role(cr.config)
    cr.run(
        args.ACTION, args.regions, step_names=args.steps, skip_steps=args.skip
    )


if __name__ == "__main__":
    main()



# Copyright 2017-2021 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import logging
import c7n.filters  # noqa - used to fix a circular import error
from c7n.tags import DEFAULT_TAG

logger = logging.getLogger(__name__)


class NotifyOnlyPolicy:
    """
    This class converts a c7n policy to a "notify only" policy. See
    :ref:`policies.notify_only` for details.

    IMPORTANT: When making changes to this class, be SURE to update
    :ref:`policies.notify_only` in the documentation.
    """

    def __init__(self, policy: dict):
        """
        Initialize a NotifyOnlyPolicy.

        :param policy: the original policy
        :type policy: dict
        """
        self._original: dict = policy
        self._mark_for_op_tags: List[str] = []
        self._fixed: dict = self._process(self._original)

    def as_notify_only(self) -> dict:
        """
        Return the policy, converted to a notify-only version.

        :return: converted policy
        :rtype: dict
        """
        return self._fixed

    def _process(self, policy: dict) -> dict:
        """
        Return the given policy, converted to notify-only.

        :param policy: the original c7n policy
        :type policy: dict
        :return: policy, modified for notify-only mode
        :rtype: dict
        """
        if 'notify_only' in policy:
            del policy['notify_only']
        for k in policy.keys():
            if k in ['comment', 'comments', 'description']:
                policy[k] = self._fix_comment(policy[k])
            if k == 'tags':
                policy[k] = self._fix_tags(policy[k])
            if k == 'actions':
                policy[k] = self._fix_actions(policy[k])
        # this needs to happen AFTER all _fix_actions calls...
        if 'filters' in policy:
            policy['filters'] = self._fix_filters(policy['filters'])
        return policy

    def _fix_filters(self, filters: List) -> List:
        """
        Given a list of filters from a policy, update them for any tagging
        changes.

        :param filters: filters from policy, or a subset thereof
        :type filters: list
        :return: fixed filters
        :rtype: list
        """
        tag_changes = {
            f'tag:{x}': f'tag:{x}-notify-only' for x in self._mark_for_op_tags
        }
        for idx, item in enumerate(filters):
            # each item should be a dict
            if not isinstance(item, type({})):
                continue
            for k in list(item.keys()):
                v = item[k]
                if isinstance(v, type([])):
                    item[k] = self._fix_filters(v)
                if k in tag_changes:
                    del item[k]
                    item[tag_changes[k]] = v
        return filters

    def _fix_comment(self, comment: str) -> str:
        """
        Convert a policy comment/comments/description to a notify only version,
        by prefixing it with the string "NOTIFY ONLY: ".

        :param comment: the original policy comment
        :type comment: str
        :return: the modified comment
        :rtype: str
        """
        return f'NOTIFY ONLY: {comment}'

    def _fix_tags(self, tags: List[str]) -> List[str]:
        """
        Convert a policy tags list to a notify only version, by appending a
        ``notify-only`` tag to the list.

        :param tags: the original tags list
        :type tags: list
        :return: the modified list, with a notify-only tag appended
        :rtype: list
        """
        return tags + ['notify-only']

    def _fix_actions(self, original: List) -> List:
        """
        Given a list of actions from a policy, return a new list of notify-only
        actions.

        * ``notify`` actions will be included unmodified
        * ``mark`` / ``tag`` actions will be passed through
          :py:meth:`~._fix_tag_action` and the result included
        * ``mark-for-op`` actions will be passed through
          :py:meth:`~._fix_mark_for_op_action` and the result included
        * ``remove-tag`` / ``unmark`` / ``untag`` actions will be passed through
          :py:meth:`~._fix_untag_action` and the result included
        * all other actions will be REMOVED

        :param original: original policy actions list
        :type original: list
        :return: new list of actions
        :rtype: list
        """
        result = []
        for item in original:
            if not isinstance(item, type({})):
                logger.info('NotifyOnlyPolicy - removing action: %s', item)
                continue
            a_type = item.get('type', '')
            if a_type == 'notify':
                result.append(self._fix_notify_action(item))
            if a_type == 'mark' or a_type == 'tag':
                result.append(self._fix_tag_action(item))
            elif a_type == 'mark-for-op':
                result.append(self._fix_mark_for_op_action(item))
            elif a_type in ['remove-tag', 'unmark', 'untag']:
                result.append(self._fix_untag_action(item))
            else:
                logger.info(
                    'NotifyOnlyPolicy - removing %s action: %s', a_type, item
                )
                continue
        return result

    def _fix_notify_action(self, item: dict) -> dict:
        """
        Fix a ``notify`` action for notify-only operation.

        If the ``violation_desc`` key is present, its value will be prefixed
        with ``NOTIFY ONLY: ``. If the ``action_desc`` key is present, its value
        will be prefixed with the string
        ``in the future (currently notify-only)``.

        :param item: the original action
        :type item: dict
        :return: the modified action
        :rtype: dict
        """
        if 'violation_desc' in item:
            item['violation_desc'] = 'NOTIFY ONLY: ' + item['violation_desc']
        if 'action_desc' in item:
            item['action_desc'] = 'in the future (currently notify-only) ' + \
                                  item['action_desc']
        return item

    def _fix_tag_action(self, item: dict) -> dict:
        """
        Fix a ``tag`` / ``mark`` action for notify-only operation.

        The string ``-notify-only`` will be appended to the ``tag`` item,
        ``key`` item, and/or every item in the ``tags`` list.

        If none of these values are set, the ``tag`` item will be set to the
        custodian ``DEFAULT_TAG``, suffixed with ``-notify-only``.

        :param item: the original action
        :type item: dict
        :return: the modified action
        :rtype: dict
        """
        if 'tag' in item:
            item['tag'] = item['tag'] + '-notify-only'
        if 'key' in item:
            item['key'] = item['key'] + '-notify-only'
        if 'tags' in item:
            item['tags'] = {
                f'{k}-notify-only': v for k, v in item['tags'].items()
            }
        if 'tag' not in item and 'key' not in item and 'tags' not in item:
            item['tag'] = f'{DEFAULT_TAG}-notify-only'
        return item

    def _fix_mark_for_op_action(self, item: dict) -> dict:
        """
        Fix a ``mark-for-op`` action for notify-only operation.

        The string ``notify-only`` will be appended to the tag name used.

        :param item: the original action
        :type item: dict
        :return: the modified action
        :rtype: dict
        """
        if 'tag' not in item:
            item['tag'] = DEFAULT_TAG
        self._mark_for_op_tags.append(item['tag'])
        item['tag'] += '-notify-only'
        return item

    def _fix_untag_action(self, item: dict) -> dict:
        """
        Fix a ``remove-tag`` / ``unmark`` / ``untag`` action for notify-only
        operation.

        All tag names in the ``tags`` list will have ``-notify-only`` appended.

        :param item: the original action
        :type item: dict
        :return: the modified action
        :rtype: dict
        """
        if 'tags' not in item:
            item['tags'] = [f'{DEFAULT_TAG}']
        item['tags'] = [f'{tag}-notify-only' for tag in item['tags']]
        return item



from setuptools import setup, find_packages
from manheim_c7n_tools.version import VERSION, PROJECT_URL

with open('README.md') as f:
    long_description = f.read()

requires = [
    'boto3',
    'botocore',
    'docutils',
    'tabulate>=0.8.0,<0.9.0',
    # In order to work with the "mu" Lambda function management tool,
    # we need PyYAML 3.x, and need it as source and not a wheel
    'pyyaml',
    'c7n==0.9.14',
    'c7n-mailer==0.6.13',
    # for building generated policy docs
    'sphinx>=1.8.0,<1.9.0',
    'sphinx_rtd_theme',
    'm2r',
    'mistune==0.8.4'
]

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Topic :: System :: Distributed Computing',
    'Topic :: System :: Systems Administration',
    'Topic :: Utilities'
]

setup(
    name='manheim-c7n-tools',
    version=VERSION,
    author='Manheim Release Engineering',
    author_email='man-releaseengineering@manheim.com',
    packages=find_packages(),
    url=PROJECT_URL,
    description='c7n policy generation script and related utilities',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requires,
    keywords="custodian aws c7n policy",
    classifiers=classifiers,
    entry_points={
        'console_scripts': [
            'policygen = manheim_c7n_tools.policygen:main',
            's3-archiver = manheim_c7n_tools.s3_archiver:main',
            'dryrun-diff = manheim_c7n_tools.dryrun_diff:main',
            'mugc = manheim_c7n_tools.vendor.mugc:main',
            'manheim-c7n-runner = manheim_c7n_tools.runner:main',
            'errorscan = manheim_c7n_tools.errorscan:main'
        ]
    }
)



# Copyright 2017-2019 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import manheim_c7n_tools.version as version

import re


class TestVersion(object):

    def test_project_url(self):
        expected = 'https://github.com/manheim/manheim-c7n-tools'
        assert version.PROJECT_URL == expected

    def test_is_semver(self):
        # see:
        # https://github.com/mojombo/semver.org/issues/59#issuecomment-57884619
        semver_ptn = re.compile(
            r'^'
            r'(?P<MAJOR>(?:'
            r'0|(?:[1-9]\d*)'
            r'))'
            r'\.'
            r'(?P<MINOR>(?:'
            r'0|(?:[1-9]\d*)'
            r'))'
            r'\.'
            r'(?P<PATCH>(?:'
            r'0|(?:[1-9]\d*)'
            r'))'
            r'(?:-(?P<prerelease>'
            r'[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*'
            r'))?'
            r'(?:\+(?P<build>'
            r'[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*'
            r'))?'
            r'$'
        )
        assert semver_ptn.match(version.VERSION) is not None



# Copyright 2017-2021 Manheim / Cox Automotive
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch, call, DEFAULT

from manheim_c7n_tools.notifyonly import NotifyOnlyPolicy
from c7n.tags import DEFAULT_TAG

pbm = 'manheim_c7n_tools.notifyonly'
pb = f'{pbm}.NotifyOnlyPolicy'


class TestInit:

    def test_simple(self):
        with patch(f'{pb}._process') as m_process:
            m_process.return_value = {'fixed': 'policy'}
            cls = NotifyOnlyPolicy({'my': 'policy'})
        assert cls._original == {'my': 'policy'}
        assert cls._fixed == {'fixed': 'policy'}
        assert cls._mark_for_op_tags == []
        assert m_process.mock_calls == [
            call({'my': 'policy'})
        ]


class NotifyOnlyTester:

    def setup(self):
        self.policy = {}
        with patch(f'{pb}._process') as m_process:
            m_process.return_value = {}
            self.cls = NotifyOnlyPolicy(self.policy)


class TestAsNotifyOnly(NotifyOnlyTester):

    def test_simple(self):
        self.cls._fixed = {'foo': 'bar'}
        assert self.cls.as_notify_only() == {'foo': 'bar'}


class TestProcess(NotifyOnlyTester):

    def test_all(self):
        policy = {
            'foo': 'bar',
            'notify_only': True,
            'baz': 'blam',
            'comment': 'commentVal',
            'comments': 'commentsVal',
            'description': 'descriptionVal',
            'tags': ['my', 'tags'],
            'actions': ['some', 'actions'],
            'filters': ['my-filters']
        }
        with patch(f'{pb}._fix_actions', autospec=True) as mock_fa:
            mock_fa.return_value = ['updated', 'actions']
            with patch(f'{pb}._fix_filters', autospec=True) as mock_ff:
                mock_ff.return_value = ['fixed', 'filters']
                res = self.cls._process(policy)
        assert res == {
            'foo': 'bar',
            'baz': 'blam',
            'comment': 'NOTIFY ONLY: commentVal',
            'comments': 'NOTIFY ONLY: commentsVal',
            'description': 'NOTIFY ONLY: descriptionVal',
            'tags': ['my', 'tags', 'notify-only'],
            'actions': ['updated', 'actions'],
            'filters': ['fixed', 'filters']
        }
        assert mock_fa.mock_calls == [
            call(self.cls, ['some', 'actions'])
        ]
        assert mock_ff.mock_calls == [
            call(self.cls, ['my-filters'])
        ]

    def test_empty_policy(self):
        policy = {
            'foo': 'bar',
            'baz': 'blam',
        }
        with patch(f'{pb}._fix_actions', autospec=True) as mock_fa:
            mock_fa.return_value = ['updated', 'actions']
            with patch(f'{pb}._fix_filters', autospec=True) as mock_ff:
                mock_ff.return_value = ['fixed', 'filters']
                res = self.cls._process(policy)
        assert res == {
            'foo': 'bar',
            'baz': 'blam',
        }
        assert mock_fa.mock_calls == []
        assert mock_ff.mock_calls == []


class TestFixFilters(NotifyOnlyTester):

    def test_simple(self):
        self.cls._mark_for_op_tags = ['tagA', 'tagB']
        filters = [
            {'tag:tagA': 'absent'},
            {'State.Name': 'running'},
            {'tag:something': 'present'},
            {'or': [
                {'and': [
                    {'something': 'else'},
                    {'tag:tagA': 'present'}
                ]},
                {'tag:tagB': 'absent'},
                {'something': 'different'},
                'not-a-dict'
            ]}
        ]
        expected = [
            {'tag:tagA-notify-only': 'absent'},
            {'State.Name': 'running'},
            {'tag:something': 'present'},
            {'or': [
                {'and': [
                    {'something': 'else'},
                    {'tag:tagA-notify-only': 'present'}
                ]},
                {'tag:tagB-notify-only': 'absent'},
                {'something': 'different'},
                'not-a-dict'
            ]}
        ]
        assert self.cls._fix_filters(filters) == expected


class TestFixComment(NotifyOnlyTester):

    def test_simple(self):
        assert self.cls._fix_comment('foo') == 'NOTIFY ONLY: foo'


class TestFixTags(NotifyOnlyTester):

    def test_simple(self):
        assert self.cls._fix_tags(['foo']) == ['foo', 'notify-only']


class TestFixActions(NotifyOnlyTester):

    def test_all(self):
        actions = [
            'something',
            {'type': 'foo'},
            {'type': 'notify'},
            {'type': 'mark'},
            {'type': 'tag'},
            {'type': 'mark-for-op'},
            {'type': 'remove-tag'},
            {'type': 'unmark'},
            {'type': 'untag'},
            {'type': 'bar'},
        ]
        with patch.multiple(
            pb,
            _fix_tag_action=DEFAULT,
            _fix_mark_for_op_action=DEFAULT,
            _fix_untag_action=DEFAULT,
            autospec=True
        ) as mocks:
            mocks['_fix_tag_action'].return_value = {'fixed': 'tag'}
            mocks['_fix_mark_for_op_action'].return_value = {'fixed': 'op'}
            mocks['_fix_untag_action'].return_value = {'fixed': 'untag'}
            res = self.cls._fix_actions(actions)
        assert res == [
            {'type': 'notify'},
            {'fixed': 'tag'},
            {'fixed': 'tag'},
            {'fixed': 'op'},
            {'fixed': 'untag'},
            {'fixed': 'untag'},
            {'fixed': 'untag'}
        ]
        assert mocks['_fix_tag_action'].mock_calls == [
            call(self.cls, {'type': 'mark'}),
            call(self.cls, {'type': 'tag'})
        ]
        assert mocks['_fix_mark_for_op_action'].mock_calls == [
            call(self.cls, {'type': 'mark-for-op'})
        ]
        assert mocks['_fix_untag_action'].mock_calls == [
            call(self.cls, {'type': 'remove-tag'}),
            call(self.cls, {'type': 'unmark'}),
            call(self.cls, {'type': 'untag'}),
        ]


class TestFixNotifyAction(NotifyOnlyTester):

    def test_no_description(self):
        policy = {
            'type': 'notify',
            'subject': 'foo'
        }
        assert self.cls._fix_notify_action(policy) == policy

    def test_description(self):
        policy = {
            'type': 'notify',
            'subject': 'foo',
            'violation_desc': 'Violation',
            'action_desc': 'actionDesc'
        }
        assert self.cls._fix_notify_action(policy) == {
            'type': 'notify',
            'subject': 'foo',
            'violation_desc': 'NOTIFY ONLY: Violation',
            'action_desc': 'in the future (currently notify-only) actionDesc'
        }


class TestFixTagAction(NotifyOnlyTester):

    def test_tag(self):
        policy = {
            'action': 'mark',
            'tag': 'foo',
            'value': 'bar'
        }
        assert self.cls._fix_tag_action(policy) == {
            'action': 'mark',
            'tag': 'foo-notify-only',
            'value': 'bar'
        }

    def test_key(self):
        policy = {
            'action': 'mark',
            'key': 'foo',
            'value': 'bar'
        }
        assert self.cls._fix_tag_action(policy) == {
            'action': 'mark',
            'key': 'foo-notify-only',
            'value': 'bar'
        }

    def test_tags(self):
        policy = {
            'action': 'mark',
            'tags': {'foo': 'bar', 'baz': 'blam'}
        }
        assert self.cls._fix_tag_action(policy) == {
            'action': 'mark',
            'tags': {'foo-notify-only': 'bar', 'baz-notify-only': 'blam'}
        }

    def test_default_tag(self):
        policy = {
            'action': 'mark',
            'value': 'bar'
        }
        assert self.cls._fix_tag_action(policy) == {
            'action': 'mark',
            'tag': f'{DEFAULT_TAG}-notify-only',
            'value': 'bar'
        }


class TestFixMarkForOpAction(NotifyOnlyTester):

    def test_tag_present(self):
        policy = {
            'action': 'mark-for-op',
            'op': 'foo',
            'tag': 'mytag',
            'days': 7
        }
        assert self.cls._mark_for_op_tags == []
        assert self.cls._fix_mark_for_op_action(policy) == {
            'action': 'mark-for-op',
            'op': 'foo',
            'tag': 'mytag-notify-only',
            'days': 7
        }
        assert self.cls._mark_for_op_tags == ['mytag']

    def test_tag_not_present(self):
        policy = {
            'action': 'mark-for-op',
            'op': 'foo',
            'days': 7
        }
        self.cls._mark_for_op_tags = ['foo', 'bar']
        assert self.cls._fix_mark_for_op_action(policy) == {
            'action': 'mark-for-op',
            'op': 'foo',
            'tag': f'{DEFAULT_TAG}-notify-only',
            'days': 7
        }
        assert self.cls._mark_for_op_tags == ['foo', 'bar', DEFAULT_TAG]


class TestFixUntagAction(NotifyOnlyTester):

    def test_simple(self):
        assert self.cls._fix_untag_action({
            'tags': ['foo', 'bar'],
            'baz': 'blam',
            'action': 'untag'
        }) == {
            'tags': ['foo-notify-only', 'bar-notify-only'],
            'baz': 'blam',
            'action': 'untag'
        }

    def test_tags_not_present(self):
        assert self.cls._fix_untag_action({
            'baz': 'blam',
            'action': 'untag'
        }) == {
            'tags': [f'{DEFAULT_TAG}-notify-only'],
            'baz': 'blam',
            'action': 'untag'
        }
